import os
import numpy as np
import h5py
from tonic.dataset import Dataset


class HSD(Dataset):
    """Heidelberg Spiking Dataset <https://arxiv.org/abs/1910.07407> contains the Spiking Heidelberg Dataset (SHD)
    and the Spiking Speech Commands dataset (SSC)."""

    base_url = "https://zenkelab.org/datasets/"
    sensor_size = (700, 1, 1)
    dtype = np.dtype([("t", int), ("x", int), ("p", int)])
    ordering = dtype.names

    def __getitem__(self, index):
        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        # adding artificial polarity of 1
        events = np.vstack(
            (
                file["spikes/times"][index],
                file["spikes/units"][index],
                np.ones(file["spikes/times"][index].shape[0]),
            )
        ).T
        # convert to microseconds
        events[:, 0] *= 1e6
        events = np.lib.recfunctions.unstructured_to_structured(events, self.dtype)
        target = file["labels"][index].astype(int)
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        return len(file["labels"])

    def _check_exists(self):
        return self._is_file_present() and self._folder_contains_at_least_n_files_of_type(
            1, ".h5"
        )


class SHD(HSD):
    """Spiking Heidelberg Digits. One of two Heidelberg Spiking Digits <https://arxiv.org/abs/1910.07407>.
    Events have (txp) ordering.
    ::

        @article{cramer2020heidelberg,
          title={The heidelberg spiking data sets for the systematic evaluation of spiking neural networks},
          author={Cramer, Benjamin and Stradmann, Yannik and Schemmel, Johannes and Zenke, Friedemann},
          journal={IEEE Transactions on Neural Networks and Learning Systems},
          year={2020},
          publisher={IEEE}
        }

    Parameters:
        save_to (string): Location to save files to on disk. Will put files in an 'hsd' subfolder.
        train (bool): If True, uses training subset, otherwise testing subset.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.

    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    test_zip = "shd_test.h5.zip"
    train_zip = "shd_train.h5.zip"
    test_md5 = "1503a5064faa34311c398fb0a1ed0a6f"
    train_md5 = "f3252aeb598ac776c1b526422d90eecb"
    folder_name = ""

    def __init__(self, save_to, train=True, transform=None, target_transform=None):
        super(HSD, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        if train:
            self.url = self.base_url + self.train_zip
            self.filename = self.train_zip
            self.file_md5 = self.train_md5
        else:
            self.url = self.base_url + self.test_zip
            self.filename = self.test_zip
            self.file_md5 = self.test_md5
        self.data_filename = self.filename[:-4]

        if not self._check_exists():
            self.download()

        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        self.classes = file["extra/keys"][()]


class SSC(HSD):
    """Spiking Speech Commands dataset. One of two Heidelberg Spiking Datasets <https://arxiv.org/abs/1910.07407>.
    Events have (txp) ordering.
    ::

        @article{cramer2020heidelberg,
          title={The heidelberg spiking data sets for the systematic evaluation of spiking neural networks},
          author={Cramer, Benjamin and Stradmann, Yannik and Schemmel, Johannes and Zenke, Friedemann},
          journal={IEEE Transactions on Neural Networks and Learning Systems},
          year={2020},
          publisher={IEEE}
        }

    Parameters:
        save_to (string): Location to save files to on disk. Will put files in an 'hsd' subfolder.
        split (string): One of 'train', 'test' or 'valid'.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.

    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    test_zip = "ssc_test.h5.zip"
    train_zip = "ssc_train.h5.zip"
    valid_zip = "ssc_valid.h5.zip"
    test_md5 = "a35ff1e9cffdd02a20eb850c17c37748"
    train_md5 = "d102be95e7144fcc0553d1f45ba94170"
    valid_md5 = "b4eee3516a4a90dd0c71a6ac23a8ae43"
    folder_name = ""

    def __init__(self, save_to, split="train", transform=None, target_transform=None):
        save_to = os.path.join(save_to, self.__class__.__name__)
        super(HSD, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        if split == "train":
            self.url = self.base_url + self.train_zip
            self.filename = self.train_zip
            self.file_md5 = self.train_md5
        elif split == "test":
            self.url = self.base_url + self.test_zip
            self.filename = self.test_zip
            self.file_md5 = self.test_md5
        elif split == "valid":
            self.url = self.base_url + self.valid_zip
            self.filename = self.valid_zip
            self.file_md5 = self.valid_md5
        self.data_filename = self.filename[:-4]

        if not self._check_exists():
            self.download()

        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        self.classes = file["extra/keys"][()]



from dataclasses import dataclass
from typing_extensions import Protocol
from typing import Any, List, Tuple
from . import functional
import numpy as np

# some slicing methods have been copied and/or adapted from
# https://gitlab.com/synsense/aermanager/-/blob/master/aermanager/preprocess.py
class Slicer(Protocol):
    def get_slice_metadata(self, data: Any) -> List[Any]:
        """
        This method returns the meta data that helps with slicing of the given data.
        Eg. it could return the indices at which the data would be sliced.
        The return value of this method should be usable by the method slice_with_meta.

        Parameters:
            data:

        Returns:
            metadata
        """
        ...

    def slice_with_metadata(self, data: Any, metadata: Any) -> List[Any]:
        """
        Slice the data using the metadata.
        Parameters:
            data:
            metadata:

        Returns:
            sliced_data
        """
        ...

    def slice(self, data: Any) -> List[Any]:
        """
        Slice the given data and return the sliced data
        Parameters:
            data:

        Returns:
            sliced_data
        """
        ...


@dataclass(frozen=True)
class SliceByTime:
    """
    Return xytp split according to fixed timewindow and overlap size
    >        <overlap>
    >|    window1     |
    >        |   window2     |

    Parameters:
        time_window: int
            Length of time for each xytp (ms)
        overlap: int
            Length of time of overlapping (ms)
        include_incomplete: bool
            include incomplete slices ie potentially the last xytp
    """

    time_window: float
    overlap: float = 0.0
    include_incomplete: bool = False

    def slice(self, data: np.ndarray) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data)
        return self.slice_with_metadata(data, metadata)

    def get_slice_metadata(self, data: np.ndarray) -> List[Tuple[int, int]]:
        t = data["t"]
        stride = self.time_window - self.overlap
        assert stride > 0

        if self.include_incomplete:
            n_slices = int(np.ceil(((t[-1] - t[0]) - self.time_window) / stride) + 1)
        else:
            n_slices = int(np.floor(((t[-1] - t[0]) - self.time_window) / stride) + 1)
        n_slices = max(n_slices, 1)  # for strides larger than recording time

        window_start_times = np.arange(n_slices) * stride + t[0]
        window_end_times = window_start_times + self.time_window
        indices_start = np.searchsorted(t, window_start_times)[:n_slices]
        indices_end = np.searchsorted(t, window_end_times)[:n_slices]
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(data: np.ndarray, metadata: List[Tuple[int, int]]):
        return [data[start:end] for start, end in metadata]


@dataclass(frozen=True)
class SliceByEventCount:
    """
    Return xytp sliced to equal number of events specified by event_count

    Parameters:
        event_count (int):  Number of events per xytp
        overlap: int
            No. of spikes overlapping in the following xytp(ms)
        include_incomplete: bool
            include incomplete slices ie potentially the last xytp
    """

    event_count: int
    overlap: int = 0
    include_incomplete: bool = False

    def slice(self, data: np.ndarray) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data)
        return self.slice_with_metadata(data, metadata)

    def get_slice_metadata(self, data: np.ndarray) -> List[Tuple[int, int]]:
        n_events = len(data)
        event_count = min(self.event_count, n_events)

        stride = self.event_count - self.overlap
        if stride <= 0:
            raise Exception("Inferred stride <= 0")

        if self.include_incomplete:
            n_slices = int(np.ceil((n_events - event_count) / stride) + 1)
        else:
            n_slices = int(np.floor((n_events - event_count) / stride) + 1)

        indices_start = (np.arange(n_slices) * stride).astype(int)
        indices_end = indices_start + event_count
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(data: np.ndarray, metadata: List[Tuple[int, int]]):
        return [data[start:end] for start, end in metadata]


@dataclass
class SliceAtIndices:
    """
    Slices data at the specified indices

    Parameters:
        start_indices: (List[Int]): List of start indices
        end_indices: (List[Int]): List of end indices (exclusive)
    """

    start_indices: np.ndarray
    end_indices: np.ndarray

    def slice(self, data: np.ndarray) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data)
        return self.slice_with_metadata(data, metadata)

    def get_slice_metadata(self, _: np.ndarray) -> List[Tuple[int, int]]:
        return list(zip(self.start_indices, self.end_indices))

    @staticmethod
    def slice_with_metadata(data: np.ndarray, metadata: List[Tuple[int, int]]):
        return [data[start:end] for start, end in metadata]


@dataclass
class SliceAtTimePoints:
    """
    Slice the data at the specified time points

    Parameters:
        tw_start: (List[Int]): List of start times
        tw_end: (List[Int]): List of end times
    """

    start_tw: np.ndarray
    end_tw: np.ndarray

    def slice(self, data: np.ndarray) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data)
        return self.slice_with_metadata(data, metadata)

    def get_slice_metadata(self, data: np.ndarray) -> List[Tuple[int, int]]:
        t = data["t"]
        indices_start = np.searchsorted(t, self.start_tw)
        indices_end = np.searchsorted(t, self.end_tw)
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(data: np.ndarray, metadata: List[Tuple[int, int]]):
        return [data[start:end] for start, end in metadata]



#! /usr/bin/env python3

import numpy as np
import math

def _gen_pixel_to_cell_mapper(h, w, cell_size):
  """
  Matrix that allows to map a pixel to a cell via indexing.
  """
  matrix = np.zeros((h, w), dtype=np.int32)
  for i in range(h):
    for j in range(w):
      # Since each frame axis is divided in cells of width _K, the cell index related to that axis can be computed by integer division x//_K.
      cell_y, cell_x = i//cell_size, j//cell_size
      # Getting the index associated to cell coordinates.
      matrix[i,j] = cell_y*math.ceil(w/cell_size) + cell_x
  return matrix
  
def _get_cell_mems(events, n_cells, n_pols, pixel_to_cell_mapper):
  """
  Generates the local memories of all cells in the frame.
  """
  cell_mems = [[[] for c in range(n_cells)] for p in range(n_pols)]
  for event in events:
    cell_mems[int(event['p'])][pixel_to_cell_mapper[event['y'], event['x']]].append(event)
  cell_mems = [[np.hstack(cell_mems[p][c]) if cell_mems[p][c] else np.empty((0,)) for c in range(n_cells)] for p in range(n_pols)]
  return cell_mems

def _get_time_surface(event, cell_mem, surface_size=10, temporal_window=500, tau=500, decay='exp'):
  """
  Accumulates the time surfaces of all the events included in the time window of the given event. 
  """
  # Empty cell.
  if cell_mem.size==0:
    return np.zeros((surface_size, surface_size))
  radius = np.array(surface_size//2).astype(cell_mem['x'].dtype)
  # Getting starting timestamp.
  t_start = max(0, event['t'] - temporal_window)
  # Computing surface coordinates for events.
  x_surface, y_surface = (event['x'] + radius - cell_mem['x']), (event['y'] + radius - cell_mem['y'])
  # Filtering events out of surface and time window.
  mask = np.asarray((x_surface>=0) & (x_surface<=2*radius) & (y_surface>=0) & (y_surface<=2*radius) & (cell_mem['t']>=t_start)).nonzero()[0]
  if mask.size==0:
    return np.zeros((surface_size, surface_size))
  # Time surfaces tensor.
  ts = np.zeros((mask.size, surface_size, surface_size), dtype=np.float32)
  # For each event, the corresponding time surface is computed using the coordinates previously calculated. 
  ts[np.arange(mask.size), y_surface[mask], x_surface[mask]] = np.exp(-(event['t'] - cell_mem['t'][mask]).astype(np.float32)/tau) if decay=='exp' else (event['t'] - cell_mem['t'][mask]).astype(np.float32)/(3*tau) + 1
  # The time surfaces are accumulated and returned in output.
  return np.sum(ts, axis=0)

def to_averaged_timesurface(
    events,
    sensor_size=(128, 128, 2),
    cell_size=10,
    surface_size=3,
    temporal_window=5e5,
    tau=5e3,
    decay="exp",
    num_workers=1
):
  """Representation that creates averaged timesurfaces for each event for one recording. Taken from the paper
    Sironi et al. 2018, HATS: Histograms of averaged time surfaces for robust event-based object classification
    https://openaccess.thecvf.com/content_cvpr_2018/papers/Sironi_HATS_Histograms_of_CVPR_2018_paper.pdf
    Parameters:
        cell_size (int): size of each square in the grid
        surface_size (int): has to be odd
        temporal_window (int): how far back to look for past events for the time averaging. Expressed in microseconds.
        tau (int): time constant to decay events around occuring event with. Expressed in microseconds.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
        merge_polarities (bool): flag that tells whether polarities should be taken into account separately or not.
        num_workers (int): number of workers to be deployed on the histograms computation. 
    Returns:
        array of histograms (numpy.Array with shape (n_cells, n_pols, surface_size, surface_size))
  """
  
  assert surface_size <= cell_size
  assert surface_size%2 != 0
  assert "x" and "y" and "t" and "p" in events.dtype.names
  assert decay=='lin' or decay=='exp'

  if num_workers>1:
    try:
      from joblib import Parallel, delayed
    except Exception as e:
      print("Error: num_workers>1 needs joblib installed.")
    use_joblib = True
  else:
    use_joblib = False

  # Getting sensor sizes, number of cells in the frame and initializing the data structures.
  width, height, n_pols = sensor_size
  n_cells = math.ceil(width/cell_size)*math.ceil(height/cell_size)
  histograms = np.zeros((n_cells, n_pols, surface_size, surface_size))
     
  # Matrix for associating an event to a cell via indexing.
  pixel_to_cell_mapper = _gen_pixel_to_cell_mapper(w=width, h=height, cell_size=cell_size)
  
  # Organizing the events in cells (local memories of HATS paper). 
  cell_mems = _get_cell_mems(events=events, n_cells=n_cells, n_pols=n_pols, pixel_to_cell_mapper=pixel_to_cell_mapper)
  
  # Time surfaces associated to each event in a cell.
  get_cell_time_surfaces = lambda cell_mem: np.stack([
    _get_time_surface(event=cell_mem[i], cell_mem=cell_mem[:i], surface_size=surface_size, temporal_window=temporal_window, tau=tau, decay=decay)
    for i in range(cell_mem.shape[0])])

  # Histogram associated to a cell, obtained by accumulatig the time surfaces of each event in the cell.
  get_cell_histogram = lambda cell: np.stack([
    np.sum(get_cell_time_surfaces(cell_mems[pol][cell]), axis=0)/max(sum([cell_mems[p][cell].shape[0] for p in range(n_pols)]), 1)
    if cell_mems[pol][cell].size>0 else
    np.zeros((surface_size, surface_size))
    for pol in range(n_pols)])

  if not use_joblib:
    histograms = np.stack([get_cell_histogram(c) for c in range(n_cells)])
  else:
    histograms = np.stack(
      Parallel(n_jobs=num_workers)(
        delayed(get_cell_histogram)(c) for c in range(n_cells)
      )
    )
  return histograms
